/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Commons", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package org.apache.commons.text.linebreak;

/**
 * This class is meant for supporting the Unicode line breaking algorithm.
 * See <a href="http://unicode.org/reports/tr14/">UTR 14</a>.
 *
 */
public class LineBreakStatus {

    private byte leftClass;
    private boolean hadSpace;

    public LineBreakStatus() {
        reset();
    }
    
    
    /**
     * Reset the status.
     * This method will reset the status to the initial state. It is meant
     * for recycling objects.
     */
    public void reset() {
        leftClass = -1;
        hadSpace = false;
    }

    /**
     * Check whether a line break may happen.
     * If the function returns "true" the line may be broken <i>before</i> the character, otherwise line
     *  breaking is prohibited. The algorithm is the table driven algorithm described in the Unicode
     *  <a href="http://unicode.org/reports/tr14/#PairBasedImplementation">technical report #14</a>.
     * The pair table is taken from @see LineBreakUtils
     * 
     * TODO: Handle CR, LF and mandatory breaks as well as better handling for AI, SA, CB and
     * other line break classes. This is not neccesary for most of the simple use cases. 
     * 
     * @param c The character.
     * @return True if a line break may happen, false if the line <i>must not</i> be broken.
     */
    public boolean nextChar(char c) {
        byte currentClass = LineBreakUtils.getLineBreakProperty(c);
        if (currentClass == LineBreakUtils.LINE_BREAK_PROPERTY_AI
            || leftClass == LineBreakUtils.LINE_BREAK_PROPERTY_XX) {
            currentClass = LineBreakUtils.LINE_BREAK_PROPERTY_AL;
        }
        if (leftClass == -1) {
            leftClass = currentClass;
            if (leftClass == LineBreakUtils.LINE_BREAK_PROPERTY_CM) {
                leftClass = LineBreakUtils.LINE_BREAK_PROPERTY_ID;
            }
            // LB 2a
            return false;
        } else if (currentClass == LineBreakUtils.LINE_BREAK_PROPERTY_SP) {
            hadSpace = true;
            return false;
        } else {
            if (currentClass == LineBreakUtils.LINE_BREAK_PROPERTY_CM && hadSpace) {
                currentClass = LineBreakUtils.LINE_BREAK_PROPERTY_ID;
            }
            byte pairProperty = LineBreakUtils.getLineBreakPairProperty(leftClass, currentClass);
            leftClass = currentClass;
            boolean savedHadSpace = hadSpace;
            hadSpace = false;
            switch (pairProperty) {
                case LineBreakUtils.PROHIBIT_BREAK :
                    return false;
                case LineBreakUtils.DIRECT_BREAK :
                    return true;
                case LineBreakUtils.INDIRECT_BREAK :
                    return savedHadSpace;
                default :
                    throw new RuntimeException("duh");
            }

        }
    }

}
