/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2003 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution, if
 *    any, must include the following acknowlegement:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowlegement may appear in the software itself,
 *    if and wherever such third-party acknowlegements normally appear.
 *
 * 4. The names "The Jakarta Project", "Commons", and "Apache Software
 *    Foundation" must not be used to endorse or promote products derived
 *    from this software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache"
 *    nor may "Apache" appear in their names without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 */
package org.apache.commons.text.spell;

import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

/**
 * This class reads MySpell compatible affix and dictionary files and creates
 * appropriate components for a spell checker. 
 *
 */
public class MyspellFilesConfiguration extends SpellCheckerConfiguration {
    private Dictionary dictionary;
    private WordComposer wordComposer;
    private SpellingCorrector spellingCorrector;

    public static final int SETSIZE = 256;
    public static final int MAXLNLEN = SETSIZE;

    private Map prefix;
    private Map suffix;
    private char tryCharacter[];
    private String encoding;
    private String compoundFlag;
    private int compoundMinimalChars = -1;
    private Map replacement;

    private class ParseContext {
        int mbr[] = new int[MAXLNLEN];
        private String filename;
        private BufferedReader b;
        private String line;
        private int lineNr;
        private StringTokenizer tok;

        private final static int TAG_TRY = 1;
        private final static int TAG_SET = 2;
        private final static int TAG_COMPOUNDFLAG = 3;
        private final static int TAG_COMPOUNDMIN = 4;
        private final static int TAG_REP = 5;
        private final static int TAG_PFX = 6;
        private final static int TAG_SFX = 7;

        ParseContext(String filename) throws IOException {
            this.filename = filename;
            b = new BufferedReader(new FileReader(filename));
            lineNr = 0;
        }

        String parseSingleValueLine(String tagname) throws SpellException {
            if (!tok.hasMoreTokens()) {
                throw new SpellException(filename + ':' + lineNr + ": missing " + tagname + " information: " + line);
            }
            String result = tok.nextToken();
            if (tok.hasMoreTokens()) {
                throw new SpellException(filename + ':' + lineNr + ": extra token: " + line);
            }
            return result;
        }

        void parseAffixFile() throws SpellException, IOException {
            suffix = new HashMap();
            prefix = new HashMap();
            Map tagmap = new HashMap();
            tagmap.put("TRY", new Integer(TAG_TRY));
            tagmap.put("SET", new Integer(TAG_SET));
            tagmap.put("COMPOUNDFLAG", new Integer(TAG_COMPOUNDFLAG));
            tagmap.put("COMPOUNDMIN", new Integer(TAG_COMPOUNDMIN));
            tagmap.put("REP", new Integer(TAG_REP));
            tagmap.put("PFX", new Integer(TAG_PFX));
            tagmap.put("SFX", new Integer(TAG_SFX));
            line = b.readLine();
            while (line != null) {
                lineNr++;
                tok = new StringTokenizer(line.trim());
                if (tok.hasMoreTokens()) {
                    Integer tag = (Integer)tagmap.get(tok.nextToken());
                    if (tag != null) {
                        switch (tag.intValue()) {
                            case TAG_TRY :
                                if (tryCharacter != null) {
                                    throw new SpellException(
                                        filename + ':' + lineNr + ": duplicate TRY strings: " + line);
                                } else {
                                    tryCharacter = parseSingleValueLine("TRY").toCharArray();
                                    break;
                                }
                            case TAG_SET :
                                if (encoding != null) {
                                    throw new SpellException(
                                        filename + ':' + lineNr + ": duplicate SET strings: " + line);
                                } else {
                                    encoding = parseSingleValueLine("SET");
                                    break;
                                }
                            case TAG_COMPOUNDFLAG :
                                if (compoundFlag != null) {
                                    throw new SpellException(
                                        filename + ':' + lineNr + ": duplicate compound flags: " + line);
                                } else {
                                    compoundFlag = parseSingleValueLine("COMPOUNDFLAG");
                                    break;
                                }
                            case TAG_COMPOUNDMIN :
                                if (compoundMinimalChars >= 0) {
                                    throw new SpellException(
                                        filename + ':' + lineNr + ": duplicate compound minimal char setting: " + line);
                                } else {
                                    compoundMinimalChars = Integer.parseInt(parseSingleValueLine("COMPOUNDMIN"));
                                    if ((compoundMinimalChars < 1) || (compoundMinimalChars > 50)) {
                                        compoundMinimalChars = 3;
                                    }
                                    break;
                                }
                            case TAG_REP :
                                if (replacement != null) {
                                    throw new SpellException(
                                        filename + ':' + lineNr + ": duplicate REP tables: " + line);
                                } else {
                                    int replacementCount = Integer.parseInt(parseSingleValueLine("REP"));
                                    if (replacementCount < 1) {
                                        throw new SpellException(
                                            filename
                                                + ':'
                                                + lineNr
                                                + ": incorrect number of entries in replacement table: "
                                                + line);
                                    }
                                    replacement = new HashMap(replacementCount);
                                    for (int j = 0; j < replacementCount; j++) {
                                        line = b.readLine();
                                        lineNr++;
                                        if (line == null) {
                                            throw new SpellException(
                                                filename
                                                    + ':'
                                                    + lineNr
                                                    + ": unexpected end of file while reading replacement table: "
                                                    + line);
                                        }
                                        tok = new StringTokenizer(line.trim());
                                        if (tok.hasMoreTokens()) {
                                            if (!tok.nextToken().equals("REP")) {
                                                throw new SpellException(
                                                    filename
                                                        + ':'
                                                        + lineNr
                                                        + ": replacement table is corrupt: "
                                                        + line);
                                            }
                                            if (tok.hasMoreTokens()) {
                                                String pattern = tok.nextToken();
                                                if (tok.hasMoreTokens()) {
                                                    replacement.put(pattern, tok.nextToken());
                                                    if (tok.hasMoreTokens()) {
                                                        throw new SpellException(
                                                            filename + ':' + lineNr + ": extra token: " + line);
                                                    }
                                                }
                                            }
                                        } else {
                                            throw new SpellException(
                                                filename + ':' + lineNr + ": missing pattern/replacement: " + line);
                                        }
                                    }
                                    break;
                                }
                            case TAG_SFX :
                                parseAffix("SFX", suffix);
                                break;
                            case TAG_PFX :
                                parseAffix("PFX", prefix);
                                break;
                            default :
                                throw new SpellException(filename + ':' + lineNr + ": unknown tag: " + line);
                        }
                    } else {
                        throw new SpellException(filename + ':' + lineNr + ": unknown tag: " + line);
                    }
                }
                line = b.readLine();
            }
        }

        private void parseAffix(String tagname, Map map) throws SpellException, IOException {
            if (!tok.hasMoreTokens()) {
                throw new SpellException(filename + ':' + lineNr + ": missing " + tagname + " information: " + line);
            }
            String affixChar = tok.nextToken();
            if (affixChar.length() > 1) {
                throw new SpellException(filename + ':' + lineNr + ": affix char too long: " + line);
            }
            String compoundString = tok.nextToken();
            if (compoundString.length() > 1 || (compoundString.charAt(0) != 'Y' && compoundString.charAt(0) != 'N')) {
                throw new SpellException(
                    filename + ':' + lineNr + ": unknown compound indicator: " + compoundString + " in: " + line);
            }
            //if (piece.charAt(0) == 'Y')
            //ff = XPRODUCT;
            boolean compoundflag = compoundString.charAt(0) == 'Y';
            int entryCount = Integer.parseInt(tok.nextToken());
            if (tok.hasMoreTokens()) {
                throw new SpellException(filename + ':' + lineNr + ": extra token: " + line);
            }
            for (int i = 0; i < entryCount; i++) {
                line = b.readLine();
                lineNr++;
                if (line == null) {
                    throw new SpellException(
                        filename + ':' + lineNr + ": unexpected end of file while reading " + tagname + ": " + line);
                }
                tok = new StringTokenizer(line.trim());
                if (tok.hasMoreTokens()) {
                    if (!tok.nextToken().equals(tagname)) {
                        throw new SpellException(
                            filename + ':' + lineNr + ": entry is corrupt, expected '" + tagname + "': " + line);
                    }
                    String localAffixChar = tok.nextToken();
                    if (!localAffixChar.equals(affixChar)) {
                        throw new SpellException(
                            filename
                                + ':'
                                + lineNr
                                + ": "
                                + tagname
                                + " entry is corrupt, expected '"
                                + affixChar
                                + "': "
                                + line);
                    }
                    AffixEntry a;
                    if (map == suffix) {
                        a = new SuffixEntry();
                    } else {
                        a = new PrefixEntry();
                    }
                    a.affix=affixChar.charAt(0);
                    a.compoundFlag = compoundflag;
                    // 3 - is string to strip or 0 for null
                    a.strip = tok.nextToken();
                    String append = tok.nextToken();
                    if (a.strip.equals("0")) {
                        a.strip = "";
                    }
                    Map it = map;
                    if (!append.equals("0")) {
                        if (map == suffix) {
                            append = new StringBuffer(append).reverse().toString();
                        }
                        int j = 0;
                        Map next = (Map)it.get(append.substring(j, j + 1));
                        j++;
                        while(next!=null && j < append.length()) {
                            it=next;
                            next = (Map)it.get(append.substring(j, j + 1));
                            j++;
                        }
                        if (next == null) {
                            j--;
                            while (j < append.length()) {
                                Map temp = new HashMap();
                                it.put(append.substring(j, j + 1), temp);
                                it = temp;
                                j++;
                            }
                        } else {
                            it=next;
                        }
                    }
                    List l = (List)it.get("");
                    if (l == null) {
                        l = new LinkedList();
                        it.put("", l);
                    }
                    l.add(a);
                    String condition = tok.nextToken();
                    encodeit(a, condition);
                    if (tok.hasMoreTokens()) {
                        throw new SpellException(filename + ':' + lineNr + ": extra token: " + line);
                    }
                }
            }
        }

        private void encodeit(AffixEntry ptr, String cs) {

            // now parse the string to create the conds array */
            int nc = cs.length();

            // if no condition just return
            if (cs.equals(".")) {
                ptr.condition = null;
                return;
            }

            ptr.condition = new int[SETSIZE];
            // now clear the conditions array */
            for (int i = 0; i < SETSIZE; i++) {
                ptr.condition[i] = 0;
            }

            int n = 0; // number of conditions
            int i = 0;
            while (i < nc) {
                char c = cs.charAt(i);
                if (c == '[') {
                    // start group
                    i++;
                    c = cs.charAt(i);
                    boolean neg = false;
                    if (c == '^') {
                        // negated
                        neg = true;
                        i++;
                        c = cs.charAt(i);
                    }
                    int nm = 0;
                    while (c != ']') {
                        mbr[nm] = c;
                        nm++;
                        i++;
                        c = cs.charAt(i);
                    }
                    if (neg) {
                        // complement so set all of them and then unset indicated ones
                        for (int j = 0; j < SETSIZE; j++) {
                            ptr.condition[j] = ptr.condition[j] | (1 << n);
                        }
                        for (int j = 0; j < nm; j++) {
                            int k = mbr[j];
                            ptr.condition[k] = ptr.condition[k] & ~(1 << n);
                        }
                    } else {
                        for (int j = 0; j < nm; j++) {
                            int k = mbr[j];
                            ptr.condition[k] = ptr.condition[k] | (1 << n);
                        }
                    }
                } else {
                    // not a group
                    if (c == '.') {
                        // wild card character so set them all
                        for (int j = 0; j < SETSIZE; j++) {
                            ptr.condition[j] = ptr.condition[j] | (1 << n);
                        }
                    } else {
                        // just set the proper bit for this char
                        ptr.condition[c] = ptr.condition[c] | (1 << n);
                    }
                }
                n++;
                i++;
            }
            ptr.numconds = n;
        }
    }

    public MyspellFilesConfiguration(String dictFileName, String affixFileName) throws IOException, SpellException {
        dictionary = new HashDictionary(dictFileName);
        ParseContext pc = new ParseContext(affixFileName);
        pc.parseAffixFile();
        wordComposer = new AffixWordComposer(prefix, suffix, dictionary);
        spellingCorrector = new SimpleSpellingCorrector();
    }

    /* (non-Javadoc)
     * @see org.apache.commons.text.spell.SpellCheckerConfiguration#getMainDictionary()
     */
    public Dictionary getMainDictionary() {
        return dictionary;
    }

    /* (non-Javadoc)
     * @see org.apache.commons.text.spell.SpellCheckerConfiguration#getWordComposer()
     */
    public WordComposer getWordComposer() {
        return wordComposer;
    }

    /* (non-Javadoc)
     * @see org.apache.commons.text.spell.SpellCheckerConfiguration#getSuggestionGenerator()
     */
    public SpellingCorrector getSpellingCorrector() {
        return spellingCorrector;
    }

}
